from pybrary.func import todo

from . import info
from .manage import Manager


class _Packager(Manager):
    def __init__(self, distro):
        super().__init__(distro)
        self.done = set()
        self.ready = False

    def _get_ready_(self):
        if self.ready: return
        self.do_init()
        self.ready = True

    def installed(self, pattern=None):
        if pattern:
            for name, ver in self.do_installed():
                if pattern in name:
                    yield name, ver
        else:
            yield from self.do_installed()

    def available(self, pattern=None):
        self._get_ready_()
        if pattern:
            for name, ver in self.do_available():
                if pattern in name.lower():
                    yield name, ver
        else:
            yield from self.do_available()

    def bigs(self):
        info('\tbigs')
        for line in self.do_bigs():
            size, pkg = line.split()
            size = int(size)
            while size>1000:
                size = size//1000
            yield f'{size:>7} {pkg}'

    def upgradable(self):
        info('\tupgradable')
        yield from self.do_upgradable()

    def update(self):
        info('\tupdate')
        self.do_update()
        for name, ver in self.upgradable():
            info(f'\t\t{name}')

    def upgrade(self):
        info('\tupgrade')
        self.do_upgrade()

    def install(self, name, ver=None):
        if name in self.done: return
        self._get_ready_()
        self.done.add(name)
        pkg = self.pkgmap.get(name, name)
        self.do_install(pkg, ver)

    def remove(self, name):
        info('\t<-- %s', name)
        pkg = self.pkgmap.get(name, name)
        self.do_remove(pkg)

    def cleanup(self):
        info('\tcleanup')
        self.do_cleanup()

    def do_init(self): todo(self)
    def do_update(self): todo(self)
    def do_upgradable(self): todo(self)
    def do_upgrade(self): todo(self)
    def do_install(self, pkg, ver=None): todo(self)
    def do_bigs(self): todo(self)
    def do_remove(self, pkg): todo(self)
    def do_cleanup(self): todo(self)
    def do_installed(self): todo(self)
    def do_available(self): todo(self)


class SystemPackager(_Packager):
    def __init__(self, distro):
        super().__init__(distro)
        self.pkgmap = distro.pkgmaps


class CommonPackager(_Packager):
    pkgmap = dict()




from inspect import cleandoc, getsource
from ast import parse, walk, Call
from textwrap import dedent

from pybrary.func import todo

from . import error


def inst(installer, installables):
    if installables:
        installables = installables.strip()
        if ' ' in installables:
            installables = installables.split()
        else:
            installables = [installables]
        for installable in installables:
            installer(installable)


class Module:
    def __init__(self, distro):
        self.distro = distro

    def __getattr__(self, attr):
        return getattr(self.distro.target, attr)

    def deploy(self, target, **kw):
        for mod in (
            c
            for c in reversed(self.__class__.mro())
            if issubclass(c, Module)
        ):
            try:
                ret = mod.do_deploy(self, target, **kw)
            except Exception as x:
                error(x)
                return False
            if not ret: return False
        return True

    def do_deploy(self, target, **kw):
        '''to be overridden
        '''
        return True

    @property
    def submodules(self):
        subs = list()
        modules = set(self.modules.items.keys())
        for meth in (self.deploy, self.do_deploy):
            tree = parse(dedent(getsource(meth)))
            for node in walk(tree):
                if isinstance(node, Call):
                    try:
                        f = node.func.attr
                        if f in ('deploy', 'do_deploy'):
                            m = node.args[0].s
                            if m in modules:
                                subs.append(m)
                    except: pass
        return subs

    def install(self, target, *, dep=None, pre=None, pkg=None, pip=None):
        inst(target.Package.install, pre)
        inst(target.deploy, dep)
        inst(target.Package.install, pkg)
        inst(target.pip.install, pip)
        return True

    @classmethod
    def help(cls):
        for mod in (
            c
            for c in cls.mro()
            if issubclass(c, Module)
        ):
            try:
                return cleandoc(mod.__doc__)
            except: pass
        return '?'



from importlib import import_module
from pkgutil import iter_modules
from functools import lru_cache

from pybrary.func import fqn
from pybrary.files import find
from pybrary.modules import load

from . import debug, error
import setux.core


@lru_cache()
def get_modules(ns):
    debug(f'{ns.__name__}')
    path, name = ns.__path__, ns.__name__ + '.'
    try:
        fil = None
        if hasattr(path, '_path'):                   # namespace
            found = list()
            for pth in path._path:
                debug(f'    {pth}')
                for fil in find(pth, r'\.py$'):
                    if fil.name=='__init__.py':
                        error(f' ! __init__ in ns {pth}')
                        return
                    nam = name+fil.stem
                    mod = load(nam, fil)
                    found.append((nam, mod))
            return found
        else:                                        # package
            return [
                (name, import_module(name))
                for finder, name, _ispkg in iter_modules(path, name)
            ]
    except Exception as x:
        error(f'\n ! {fil or name}\n ! {x}\n')


def get_raw_plugins(ns, cls):
    for mod, module in get_modules(ns):
        for plg, plugin in module.__dict__.items():
            try:
                if issubclass(plugin, cls):
                    yield mod, plg, plugin
            except TypeError: pass


def get_plugins(ns, cls):
    get_core = get_raw_plugins(setux.core, cls)
    core = set(cls for _m, _p, cls in get_core)
    return [
        (mod, plg, plugin)
        for mod, plg, plugin in get_raw_plugins(ns, cls)
        if plugin not in core
    ]


class Plugins:
    def __init__(self, distro, Base, ns):
        self.distro = distro
        self.Base = Base
        self.ns = ns
        self.items = dict()
        self.collect()

    def parse(self, mod, plg, plugin):
        return plg, plugin

    def sort(self):
        ''' sort self.items in place
        '''

    def collect(self):
        plugins = get_plugins(self.ns, self.Base)
        for mod, plg, plugin in plugins:
            key, val = self.parse(mod, plg, plugin)
            if key and val:
                self.items[key] = val
        self.sort()
        for mod in self.items.values():
            debug('%s registred', fqn(mod))


class Distros(Plugins):
    def sort(self):
        def sort_key(item):
            name, cls = item
            return len(cls.distro_lineage(cls))

        self.items = dict(
            item
            for item in reversed(sorted(
                self.items.items(),
                key = sort_key
            ))
        )


class DistroPlugin(Plugins):
    def parse(self, mod, plg, plugin):
        lineage = self.distro.lineage
        if plg in lineage:
            name = '.'.join(mod.split('.')[2:])
            old = self.items.get(mod)
            if old:
                idx = lineage.index(plg)
                old_idx = lineage.index(old.__name__)
                if idx > old_idx:
                    return name, plugin
            else:
                return name, plugin
        return None, None


class Modules(DistroPlugin): pass


class Managers(DistroPlugin): pass





from shlex import split
from subprocess import (
    PIPE,
    CalledProcessError,
    run,
)

from pybrary.func import todo

from .errors import (
    MissingModuleError,
    ModuleTypeError,
    UnsupportedDistroError,
    ExecError,
)
from . import debug, info, error
from .distro import Distro
from .module import Module
from . import plugins
import setux.distros


# pylint: disable= filter-builtin-not-iterating


class Target:
    def __init__(self, *,
        name = None,
        distro = None,
        sudo = None,
        outdir = None,
        exclude = None,
    ):
        self.name = name or 'target'
        self.outdir = outdir
        self.sudo = sudo
        self.release_infos = None

        self.cnx = self.chk_cnx()
        if self.cnx:
            self.distros = plugins.Distros(self,
                Distro, setux.distros
            )
            self.probe_distro()
            self.sudo = self.distro.login.id != 0
            self.exclude = exclude
        else:
            self.distro = None

    @property
    def outdir(self):
        return getattr(self, '_outdir_', None)

    @outdir.setter
    def outdir(self, path):
        self._outdir_ = path
        if path:
            run(f'mkdir -p {path}', shell=True)
            self.set_trace()

    def chk_cnx(self, report='quiet'):
        ''' to be overwriten
        '''
        return True

    def probe_distro(self):
        Distros = self.distros.items.values()
        infos = Distro.release_default(self)
        for Dist in Distros:
            if Dist.release_check(self, infos):
                self.distro = Dist(self)
                debug(f'{self.distro.host.name} : {self.distro.name}')
                break
        else:
            raise UnsupportedDistroError(self)

    def set_trace(self):
        self.outrun = f'{self.outdir}/{self.name}.run'
        self.outlog = f'{self.outdir}/{self.name}.log'
        try:
            with open(self.outrun, 'w') as log:
                debug(f'outrun : {log.name}')
            with open(self.outlog, 'w') as log:
                debug(f'outlog : {log.name}')
        except Exception as x:
            error(x)
            self.outdir = None

    def trace(self, cmd, ret, out, err, **kw):
        if self.outdir:
            with open(self.outrun, 'a') as log:
                log.write(f'{cmd}\n')

            with open(self.outlog, 'a') as log:
                log.write(f'\n[{ret:^3}] {cmd}\n')
                if out:
                    if kw.get('report')=='quiet':
                        log.write(f'[out] ... ({len(out)})\n')
                    else:
                        if len(out)==1:
                            out = out[0]
                        else:
                            out = '\n'+'\n'.join(out)
                        log.write(f'[out] {out}\n')
                if err:
                    if len(err)==1:
                        err = err[0]
                    else:
                        err = '\n'+'\n'.join(err)
                    log.write(f'[err] {err}\n')

    def __getattr__(self, attr):
        return getattr(self.distro, attr)

    def parse(self, *arg, **kw):
        shell = kw.get('shell')
        if shell is None and len(arg)==1:
            shell = any(x in arg[0] for x in '*|>?<')
            kw['shell'] = shell

        args = []

        if kw.pop('sudo', True) and self.sudo:
            args.append('sudo')

        if not shell and len(arg)==1:
            arg = filter(None,
                (i.strip() for i in split(arg[0]))
            )
        args.extend(arg)
        return args, kw

    def run(self, *arg, report='normal', critical=True, raw=False, skip=None, **kw):
        def log(*msg):
            if report=='verbose':
                debug(*msg)

        cmd = arg
        command = ' '.join(cmd)
        if kw.get('shell'):
            cmd = command

        try:
            # log('running "%s" ...', command)
            try:
                proc = run(cmd, stdout=PIPE, stderr=PIPE, **kw)
            except OSError:
                kw['shell'] = True
                proc = run(cmd, stdout=PIPE, stderr=PIPE, **kw)

            out = proc.stdout.decode("utf-8").strip()
            if out:
                if report!='quiet':
                    debug("%s [out]:\n%s", command, out)
                if not raw:
                    out = [i.strip() for i in out.split('\n')]
                    if skip:
                        out = [i for i in out if not skip(i)]

            err = proc.stderr.decode("utf-8").strip()
            if err:
                log("%s [err]:\n%s", command, err)
                if not raw:
                    err = [i.strip() for i in err.split('\n')]
                    if skip:
                        err = [i for i in err if not skip(i)]

            ret = proc.returncode
            log('"%s" [ret]: %s', command, ret)

            return ret, out, err

        except CalledProcessError as exc:
            if critical:
                error(
                    "\n%s > %s\n%s\n%s",
                    " ".join(exc.cmd),
                    exc.returncode,
                    exc.stdout.decode("utf-8"),
                    exc.stderr.decode("utf-8"),
                )
            return -1, "ERROR", str(exc)

        except Exception as exc:
            error("%s raised: %s", cmd, exc)
            raise

    def deploy(self, module, **kw):
        if not self.cnx:
            error('deploy ! no cnx')
            return

        if isinstance(module, str):
            try:
                cls = self.distro.modules.items[module]
            except KeyError:
                raise MissingModuleError(module, self.distro)
        else:
            cls = module

        try:
            if not issubclass(cls, Module):
                raise ModuleTypeError(cls)
        except TypeError:
            raise ModuleTypeError(cls)

        report = kw.pop('report', 'normal') != 'quiet'
        ret = cls(self.distro).deploy(self, **kw)
        if report:
            params = ', '.join(f'{k}={v}' for k, v in kw.items()) if kw else ''
            status = '.' if ret else 'X'
            info(f'\tdeploy {module} {params} {status}')
        return ret

    def rsync_opt(self):
        ''' additional rsync opts
        '''

    def rsync(self, *arg, **kw):
        ret, out, err =  Target.run(self, 'rsync --version', report='quiet')
        if ret: self.Package.install('rsync')
        kw['sudo'] = False
        arg, kw = self.parse(*arg, **kw)
        cmd = 'rsync -qlpogcr -zz --delete'.split()
        if self.exclude:
            cmd.extend(['--exclude-from', self.exclude, '--delete-excluded'])
        opt = self.rsync_opt()
        if opt:
            cmd.append(opt)
        cmd.extend(arg)
        kw['report'] = 'verbose'
        ret, out, err =  Target.run(self, *cmd, **kw)
        self.trace('rsync '+' '.join(arg), ret, out, err, **kw)
        return ret, out, err

    def script(self, content, cmd=None, user=None, path=None, name=None, trim=True, remove=True, report='quiet'):
        path = path or '/tmp/setux'
        self.run(f'mkdir -p {path}')
        self.run(f'chmod 777 {path}')
        name = name or 'script'
        full = '/'.join((path, name))
        if trim:
            lines = (line.strip() for line in content.split('\n'))
            content = '\n'.join(line for line in lines if line)+'\n'
        self.write(full, content, report='quiet')
        if cmd:
            if user:
                cmd = f'sudo -u {user} {cmd}'
            self.run(f'chmod 644 {full}')
            ret, out, err = self.run(cmd.format(full))
        else:
            self.run(f'chmod +x {full}')
            ret, out, err = self.run(full)
        if remove:
            self.run(f'rm {full}', report='quiet')
        return ret, out, err

    def read(self, path, mode='rt', report='normal'): todo(self)
    def write(self, path, content, mode='wt', report='normal'): todo(self)
    def send(self, local, remote): todo(self)
    def fetch(self, remote, local): todo(self)
    def sync(self, src, dst=None): todo(self)
    def export(self, path): todo(self)
    def remote(self, module, export_path=None, **kw): todo(self)




from time import sleep

from pybrary.func import todo

from . import info
from .manage import Manager


# pylint: disable=assignment-from-no-return


class Service(Manager):
    def __init__(self, distro):
        super().__init__(distro)
        self.svcmap = distro.svcmap

    def status(self, name):
        svc = self.svcmap.get(name, name)
        up = self.do_status(svc)
        info(f'\tservice {name} {"." if up else "X"}')
        return up

    def wait(self, name, up=True):
        sleep(1)
        for _ in range(3):
            if self.status(name) is up: break
            sleep(3)

    def enable(self, name):
        svc = self.svcmap.get(name, name)
        if not self.do_enabled(svc):
            info(f'\tenable {name}')
            self.do_enable(svc)
            enabled = self.do_enabled(svc)
            info(f'\t{name} enabled {"." if enabled else "X"}')

    def disable(self, name):
        svc = self.svcmap.get(name, name)
        if self.do_enabled(svc):
            info(f'\tdisable {name}')
            self.do_disable(svc)
            enabled = self.do_enabled(svc)
            info(f'\t{name} disabled {"." if not enabled else "X"}')

    def start(self, name):
        svc = self.svcmap.get(name, name)
        if not self.status(svc):
            info(f'\tstart {name}')
            self.do_start(svc)
            self.wait(name)

    def stop(self, name):
        svc = self.svcmap.get(name, name)
        if self.status(svc):
            info(f'\tstop {name}')
            self.do_stop(svc)
            self.wait(name, up=False)

    def restart(self, name):
        svc = self.svcmap.get(name, name)
        if self.status(svc):
            info(f'\trestart {name}')
            self.do_restart(svc)
            self.wait(name)
        else:
            self.start(name)

    def do_enabled(self, svc): todo(self)
    def do_status(self, svc): todo(self)
    def do_enable(self, svc): todo(self)
    def do_disable(self, svc): todo(self)
    def do_start(self, svc): todo(self)
    def do_stop(self, svc): todo(self)
    def do_restart(self, svc): todo(self)



from . import info


# pylint: disable=no-member


class Manager:
    def __init__(self, distro, quiet=False):
        self.distro = distro
        self.target = distro.target
        self.run = self.target.run
        self.key = None
        self.quiet = quiet

    def __str__(self):
        base = self.__class__.__bases__[0].__name__
        return f'{base}.{self.manager}'


class Deployable(Manager):
    def __call__(self, key, *args, **spec):
        self.key = key
        self.args = args
        self.spec = {
            k : str(v)
            for k, v in spec.items()
        }
        return self

    def deploy(self, msg=''):
        if msg: msg = f'{msg}:\n'
        status = f'{"." if self.set() else "X"}'
        if not self.quiet:
            info(f'{msg}\t{self.manager} {self.key} {status}')
        return status=='.'

    def __str__(self):
        return f'{self.__class__.__name__} {self.get()}'


class SpecChecker(Deployable):
    def chk(self, name, value, spec):
        return value == spec

    def check(self):
        data = self.get()
        if data:
            for k, v in self.spec.items():
                # if data.get(k) != v:
                if not self.chk(k, data.get(k), v):
                    return False       # mismatch
            return True                # conform
        return None                    # absent

    def set(self):
        ok = self.check()
        if ok:
            return ok
        else:
            if ok is None:
                self.cre()
            data = self.get()
            if not data: return None
            for k, v in self.spec.items():
                if not self.chk(k, data.get(k), v):
                    self.mod(k, v)
                    data = self.get()
            return self.check()


class ArgsChecker(Deployable):
    def check(self):
        data = self.get()
        if data:
            for arg in self.args:
                if arg not in data:
                    return False       # mismatch
            return True                # conform
        return None                    # absent

    def set(self):
        data = self.get()
        for arg in data:
            if arg not in self.args:
                self.rm(arg)
        for arg in self.args:
            if arg not in data:
                self.add(arg)
        return self.check()



class SetuxError(Exception):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return f'\n\n    ! {self.msg} !\n\n'


class MissingModuleError(SetuxError):
    def __init__(self, module, distro):
        lineage = '.'.join(distro.lineage)
        super().__init__(
            f'module "{module}" not defined for "{lineage}"'
        )


class ModuleTypeError(SetuxError):
    def __init__(self, module):
        super().__init__(
            f'"{module}" must be a module, not [{type(module).__name__}]'
        )


class UnsupportedDistroError(SetuxError):
    def __init__(self, target):
        super().__init__(
            f'Unsupported distro on {target}'
        )


class ExecError(SetuxError):
    def __init__(self, cmd, ret, out, err):
        sep = '\n    > '
        out = sep+sep.join(out) if out else ''
        sep = '\n    X '
        err = sep+sep.join(err) if err else ''
        super().__init__(
            f'{cmd} ! {ret} ! {out} {err}'
        )



from os.path import join, dirname, abspath

from setuptools import setup, find_namespace_packages

from setux.core import __version__

curdir = abspath(dirname(__file__))
readme = open(join(curdir, 'README.rst')).read()

setup(
    name             = 'setux_core',
    version          = __version__,
    description      = 'System deployment',
    long_description = readme,
    keywords         = ['utility', ],
    url              = 'https://github.com/louis-riviere-xyz/setux_core',
    author           = 'Louis RIVIERE',
    author_email     = 'louis@riviere.xyz',
    license          = 'MIT',
    classifiers      = [
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    install_requires = [
        'pybrary>=0.20.38.0',
    ],
    packages = find_namespace_packages(
        include=['setux.*']
    ),
)



from pybrary.func import memo

from . import debug
from .manage import Manager
from .module import Module
from . import plugins
import setux.managers
import setux.modules


# pylint: disable=bad-staticmethod-argument


class Distro:
    Package = None
    pkgmap = dict()
    Service = None
    svcmap = dict()

    def __init__(self, target):
        self.name = self.__class__.__name__
        self.target = target
        self.managers = plugins.Managers(self,
            Manager, setux.managers
        )
        self.modules = plugins.Modules(self,
            Module, setux.modules
        )
        self.set_managers()

    def __str__(self):
        return f'Distro : {self.name}'

    def set_managers(self):
        todo = list(self.managers.items.values())
        while todo:
            for manager in list(todo):
                todo.remove(manager)
                if manager.manager==self.Package:
                    self.Package = manager(self)
                    debug('%s Package %s', self.name, manager.manager)
                elif manager.manager==self.Service:
                    self.Service = manager(self)
                    debug('%s Service %s', self.name, manager.manager)
                else:
                    setattr(self, manager.manager, manager(self))
                    debug('%s %s %s', self.name, manager.manager, '.')

    @classmethod
    def release_default(cls, target):
        if target.release_infos is None:
            ret, out, err = target.run('cat /etc/*-release', report='quiet', shell=True)
            target.release_infos = dict(l.split('=') for l in out if '=' in l)
            debug('%s %s', target, target.release_infos)
        return target.release_infos

    @classmethod
    def release_name(cls, infos):
        did = infos['ID'].strip()
        ver = infos['VERSION_ID'].strip()
        return f'{did}_{ver}'

    @classmethod
    def release_check(cls, target, infos=None):
        if hasattr(cls, 'release_infos'):
            try:
                infos = cls.release_infos(target)
            except: pass
        if not infos:
            infos = cls.release_default(target)
        try:
            return cls.release_name(infos) == cls.__name__
        except: return False

    @staticmethod
    def distro_bases(cls):
        return list(reversed([
            base
            for base in cls.__mro__
            if issubclass(cls, Distro)
        ]))[1:]

    @memo
    def bases(self):
        return Distro.distro_bases(self.__class__)

    @staticmethod
    def distro_lineage(cls):
        return [b.__name__ for b in Distro.distro_bases(cls)]

    @memo
    def lineage(self):
        return [b.__name__ for b in self.bases]

    @memo
    def pkgmaps(self):
        pkgs = dict()
        for distro in self.bases:
            pkgs.update(distro.pkgmap)
        return pkgs

    @memo
    def svcmaps(self):
        svcs = dict()
        for distro in self.bases:
            svcs.update(distro.svcmap)
        return svcs
